"""Manages all Railway-specific aspects of the deployment process.

Notes:
- ...
"""

import django_simple_deploy

from dsd_railway.platform_deployer import PlatformDeployer
from .plugin_config import plugin_config
from .cli import PluginCLI, validate_cli


@django_simple_deploy.hookimpl
def dsd_get_plugin_config():
    """Get platform-specific attributes needed by core."""
    return plugin_config


@django_simple_deploy.hookimpl
def dsd_get_plugin_cli(parser):
    """Get plugin's CLI extension."""
    plugin_cli = PluginCLI(parser)


@django_simple_deploy.hookimpl
def dsd_validate_cli(options):
    """Validate and parse plugin-specific CLI args."""
    validate_cli(options)


@django_simple_deploy.hookimpl
def dsd_deploy():
    """Carry out platform-specific deployment steps."""
    platform_deployer = PlatformDeployer()
    platform_deployer.deploy()



"""A collection of messages used during the configuration and deployment process."""

# For conventions, see documentation in core deploy_messages.py

from textwrap import dedent

from django.conf import settings


confirm_automate_all = """
The --automate-all flag means django-simple-deploy will:
- ...
- Commit all changes to your project that are necessary for deployment.
- Push these changes to Railway.
- Open your deployed project in a new browser tab.
"""

cancel_railway = """
Okay, cancelling Railway configuration and deployment.
"""

# DEV: This could be moved to deploy_messages, with an arg for platform and URL.
cli_not_installed = """
In order to deploy to Railway, you need to install the Railway CLI.
  See here: ...
After installing the CLI, you can run the deploy command again.
"""

cli_logged_out = """
You are currently logged out of the Railway CLI. Please log in,
  and then run the deploy command again.
You can log in from  the command line:
  $ ...
"""


# --- Dynamic strings ---
# These need to be generated in functions, to display information that's determined as
# the script runs.

def success_msg(log_output=""):
    """Success message, for configuration-only run.

    Note: This is immensely helpful; I use it just about every time I do a
      manual test run.
    """

    msg = dedent(
        f"""
        --- Your project is now configured for deployment on Railway ---

        To deploy your project, you will need to:
        - Commit the changes made in the configuration process.
            $ git status
            $ git add .
            $ git commit -am "Configured project for deployment."
        - Push your project to Railway's servers:
            $ ...
        - Open your project:
            $ ...    
        - As you develop your project further:
            - Make local changes
            - Commit your local changes
            - Run `...` again to push your changes.
    """
    )

    if log_output:
        msg += dedent(
            f"""
        - You can find a full record of this configuration in the dsd_logs directory.
        """
        )

    return msg


def success_msg_automate_all(deployed_url, project_id):
    """Success message, when using --automate-all."""

    msg = dedent(
        f"""

        --- Your project should now be deployed on Railway ---

        It should have opened up in a new browser tab. If you see a
          "server not available" message, wait a minute or two and
          refresh the tab. It sometimes takes a few minutes for the
          server to be ready.
        - You can also visit your project at {deployed_url}
        - Project ID: {project_id}

        If you make further changes and want to push them to Railway,
        commit your changes and then run `...`.
    """
    )
    return msg



"""Manages all Railway-specific aspects of the deployment process.

Notes:

Add a new file to the user's project, using a template:

    def _add_dockerfile(self):
        # Add a minimal dockerfile.
        template_path = self.templates_path / "dockerfile_example"
        context = {
            "django_project_name": dsd_config.local_project_name,
        }
        contents = plugin_utils.get_template_string(template_path, context)
"""

import sys, os, re, json
import subprocess
import time
import webbrowser
from pathlib import Path

from django.utils.safestring import mark_safe

import requests

from . import deploy_messages as platform_msgs
from .plugin_config import plugin_config

from django_simple_deploy.management.commands.utils import plugin_utils
from django_simple_deploy.management.commands.utils.plugin_utils import dsd_config
from django_simple_deploy.management.commands.utils.command_errors import (
    DSDCommandError,
)


class PlatformDeployer:
    """Perform the initial deployment to Railway

    If --automate-all is used, carry out an actual deployment.
    If not, do all configuration work so the user only has to commit changes, and ...
    """

    def __init__(self):
        self.templates_path = Path(__file__).parent / "templates"

    # --- Public methods ---

    def deploy(self, *args, **options):
        """Coordinate the overall configuration and deployment."""
        plugin_utils.write_output("\nConfiguring project for deployment to Railway...")

        self._validate_platform()
        self._prep_automate_all()

        # Configure project for deployment to Railway
        self._modify_settings()
        self._make_static_dir()
        self._add_requirements()

        self._conclude_automate_all()
        self._show_success_message()

    # --- Helper methods for deploy() ---

    def _validate_platform(self):
        """Make sure the local environment and project supports deployment to Railway.

        Returns:
            None
        Raises:
            DSDCommandError: If we find any reason deployment won't work.
        """
        # Use local project name for deployed project, if no custom name passed.
        if not dsd_config.deployed_project_name:
            dsd_config.deployed_project_name = dsd_config.local_project_name

    def _prep_automate_all(self):
        """Take any further actions needed if using automate_all."""
        pass



    def _modify_settings(self):
        """Add Railway-specific settings."""
        msg = "\nAdding a Railway-specific settings block."
        plugin_utils.write_output(msg)

        template_path = self.templates_path / "settings.py"
        plugin_utils.modify_settings_file(template_path)

    def _make_static_dir(self):
        """Add a static/ dir if needed."""
        msg = "\nAdding a static/ directory and a placeholder text file."
        plugin_utils.write_output(msg)

        path_static = Path("static")
        plugin_utils.add_dir(path_static)

        # Write a placeholder file, to be picked up by Git.
        path_placeholder = path_static / "placeholder.txt"
        contents = "Placeholder file, to be picked up by Git.\n"
        plugin_utils.add_file(path_placeholder, contents)

    def _add_requirements(self):
        """Add requirements for deploying to Railway."""
        requirements = [
            "gunicorn",
            "whitenoise",
            "psycopg",
            "psycopg-binary",
            "psycopg-pool",
        ]
        plugin_utils.add_packages(requirements)

    def _conclude_automate_all(self):
        """Finish automating the push to Railway.

        - Commit all changes.
        - ...
        """
        # Making this check here lets deploy() be cleaner.
        if not dsd_config.automate_all:
            return

        plugin_utils.commit_changes()

        # Initialize empty project on Railway.
        plugin_utils.write_output("  Initializing empty project on Railway...")
        cmd = f"railway init --name {dsd_config.deployed_project_name}"
        plugin_utils.run_slow_command(cmd)

        # Deploy the project.
        msg = "  Pushing code to Railway."
        msg += "\n  You'll see a database error, which will be addressed in the next step."
        plugin_utils.write_output(msg)
        
        cmd = "railway up"
        try:
            plugin_utils.run_slow_command(cmd)
        except subprocess.CalledProcessError:
            msg = "  Expected error, because no Postgres database exists yet. Continuing deployment."
            plugin_utils.write_output(msg)

        # Add a database.
        msg = "  Adding a database..."
        plugin_utils.write_output(msg)

        cmd = "railway add --database postgres"
        output = plugin_utils.run_quick_command(cmd)
        plugin_utils.write_output(output)

        # Set env vars.
        self._set_env_vars()

        # Redeploy.
        cmd = f"railway redeploy --service {dsd_config.deployed_project_name} --yes"
        output = plugin_utils.run_quick_command(cmd)
        plugin_utils.write_output(output)

        # Generate a Railway domain.
        msg = "  Generating a Railway domain..."
        plugin_utils.write_output(msg)
        cmd = f"railway domain --port 8080 --service {dsd_config.deployed_project_name} --json"
        output = plugin_utils.run_quick_command(cmd)

        output_json = json.loads(output.stdout.decode())
        self.deployed_url = output_json["domain"]

        # Get project ID.
        msg = "  Getting project ID..."
        plugin_utils.write_output(msg)
        cmd = "railway status --json"
        output = plugin_utils.run_quick_command(cmd)

        output_json = json.loads(output.stdout.decode())
        plugin_config.project_id = output_json["id"]

        # Wait {pause} before opening.
        pause = 20
        msg = f"  Waiting {pause}s for deployment to finish..."
        plugin_utils.write_output(msg)

        # Wait for a 200 response.
        pause = 10
        timeout = 300
        for _ in range(int(timeout/pause)):
            msg = "  Checking if deployment is ready..."
            plugin_utils.write_output(msg)
            r = requests.get(self.deployed_url)
            if r.status_code == 200:
                break

            time.sleep(pause)

        webbrowser.open(self.deployed_url)

        msg = f"  If you get an error page, refresh the browser in a minute or two."
        plugin_utils.write_output(msg)


    def _show_success_message(self):
        """After a successful run, show a message about what to do next.

        Describe ongoing approach of commit, push, migrate.
        """
        if dsd_config.automate_all:
            msg = platform_msgs.success_msg_automate_all(self.deployed_url, plugin_config.project_id)
        else:
            msg = platform_msgs.success_msg(log_output=dsd_config.log_output)
        plugin_utils.write_output(msg)

    
    def _set_env_vars(self):
        """Set required environment variables for Railway."""
        msg = "  Setting environment variables on Railway..."
        plugin_utils.write_output(msg)

        env_vars = [
            '--set "PGDATABASE=${{Postgres.PGDATABASE}}"',
            '--set "PGUSER=${{Postgres.PGUSER}}"',
            '--set "PGPASSWORD=${{Postgres.PGPASSWORD}}"',
            '--set "PGHOST=${{Postgres.PGHOST}}"',
            '--set "PGPORT=${{Postgres.PGPORT}}"',
        ]

        cmd = f"railway variables {' '.join(env_vars)} --service {dsd_config.deployed_project_name}"
        output = plugin_utils.run_quick_command(cmd)
        plugin_utils.write_output(output)



"""Extends the core django-simple-deploy CLI."""

import json
import shlex
import subprocess

from django_simple_deploy.management.commands.utils.plugin_utils import dsd_config
from django_simple_deploy.management.commands.utils.command_errors import (
    DSDCommandError,
)

from .plugin_config import plugin_config


class PluginCLI:

    def __init__(self, parser):
        """Add plugin-specific args."""
        group_desc = "Plugin-specific CLI args for dsd-railway"
        plugin_group = parser.add_argument_group(
            title="Options for dsd-railway",
            description = group_desc,
        )

        # plugin_group.add_argument(
        #     "--vm-size",
        #     type=str,
        #     help="Name for a preset vm-size configuration, ie `shared-cpu-2x`.",
        #     default="",
        # )


def validate_cli(options):
    """Validate options that were passed to CLI."""

    # vm_size = options["vm_size"]
    # _validate_vm_size(vm_size)

    pass


# --- Helper functions ---

# def _validate_vm_size(vm_size):
#     """Validate the vm size arg that was passed."""
#     if not vm_size:
#         return

#     if not dsd_config.unit_testing:
#         cmd = "fly platform vm-sizes --json"
#         cmd_parts = shlex.split(cmd)
#         output = subprocess.run(cmd_parts, capture_output=True)
#         allowed_sizes = list(json.loads(output.stdout).keys())
#     else:
#         allowed_sizes = ["shared-cup-1x", "shared-cpu-2x"]

#     if vm_size not in allowed_sizes:
#         msg = f"The vm-size {vm_size} requested is not available."
#         msg += f"\n  Allowed sizes: {' '.join(allowed_sizes)}"
#         raise DSDCommandError(msg)

#     # vm_size is valid. Set the relevant plugin_config attribute.
#     plugin_config.vm_size = vm_size


{{current_settings}}

# Railway settings.
import os

if os.environ.get("RAILWAY_PROJECT_NAME", ""):
    DEBUG = False

    # Configure a Postgres db.
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql',
            'NAME': os.environ.get("PGDATABASE", ""),
            'USER': os.environ.get("PGUSER", ""),
            'PASSWORD': os.environ.get("PGPASSWORD", ""),
            'HOST': os.environ.get("PGHOST", ""),
            'PORT': os.environ.get("PGPORT", ""),
        }
    }

    # Static files config.
    STATIC_URL = 'static/'
    STATIC_ROOT = os.path.join(BASE_DIR, "staticfiles")
    STATICFILES_DIRS = [os.path.join(BASE_DIR, "static")]

    MIDDLEWARE.insert(1, 'whitenoise.middleware.WhiteNoiseMiddleware')

    # Allowed hosts, CSRF.
    ALLOWED_HOSTS = [".up.railway.app"]
    CSRF_TRUSTED_ORIGINS = ['https://*.railway.app']
